package edu.assistment.experiments.inclinedplane.controller;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import edu.assistment.experiments.inclinedplane.model.InclinedPlaneModel;
import edu.assistment.experiments.inclinedplane.simulation.InclinedPlaneSimulationEngine;
import edu.assistment.simulation.model.SimulationThread;
import edu.assistment.simulation.view.SimulationControlJPanel;

/**
 * Handles events from the 'run' and 'reset' button,
 * causing the simulation to run or reset.
 * 
 * @author Mike Sao Pedro (mikesp@wpi.edu)
 */
public class SimulationController implements ActionListener
{
	private SimulationControlJPanel view;
	private InclinedPlaneSimulationEngine engine;
	private InclinedPlaneModel model;
	private SimulationThread simThread;
	
	public SimulationController(SimulationControlJPanel sim_panel, InclinedPlaneModel ip_model)
	{
		view = sim_panel;
		model = ip_model;
		simThread = null;
		
		view.getRunButton().addActionListener(this);
		view.getResetButton().addActionListener(this);
		engine = new InclinedPlaneSimulationEngine(model);
	}
	
	/**
	 * @return true if the simulation has stopped and results are available
	 */
	public boolean resultsAvailable()
	{
		return !isRunning();
	}
	
	/**
	 * Determine if the simulation thread is running
	 * 
	 */
	public boolean isRunning()
	{
		return simThread != null && simThread.isRunning();
	}
	
	public void actionPerformed(ActionEvent e) 
	{
		if(e.getSource() == view.getRunButton())
		{
			runExperiment();
		}
		else if(e.getSource() == view.getResetButton())
		{
			resetExperiment();
		}
		else
		{
			throw new IllegalArgumentException("Received button from unknown source! " + e.getSource().toString());
		}
		
	}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
	
	public void runExperiment()
	{
		simThread = new SimulationThread(model, engine);
		Thread t = new Thread(simThread);
		t.start();
	}
	
	public void resetExperiment()
	{
		if(simThread != null)
		{
			simThread.stopSimulation();
		}
		
		// --- Reset all the dynamic parts of the model
		model.setCurrentTime(0.0);
		model.setBallAgainstRamp();
		model.fireModelChangedEvent(null);
	}

}
