package edu.assistment.experiments.pedagogy;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.HashMap;
import java.util.List;

import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JTable;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.border.LineBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import edu.assistment.experiments.inclinedplane.controller.SimulationController;
import edu.assistment.experiments.inclinedplane.model.InclinedPlaneModel;
import edu.assistment.experiments.inclinedplane.view.InclinedPlaneExperimentComboBoxJPanel;
import edu.wpi.cetask.Plan;
import edu.wpi.cetask.Task;

/**
 * Notes taken during lecture:
 * ===========================================================================
 * - Think of GUI actions as doing "primitive" actions
 * 
 * - When 'agent' does something, you need a pathway from javascript to
 *   the GUI elements
 *     - Send a variable to javascript representing the appropriate java-side 
 *       "model" stuff
 *     
 * - When 'users' do something, you need a pathway from the GUI change to
 *   the task model (thru listeners)
 * 
 * - Guide
 *    - public done method
 *       - constructs a particular task type w/ args
 *          - task "D15" 
 * 
 * - USE REFLECTION to do the pretty text print stuff.
 * 
 * - Try to have 'available options' rather than doing a 'script'.
 * 
 * - Requirement: don't worry about speech until the end.
 * 
 * ============================================================
 * 
 * - Class in java called JSObject (JavaScript object)
 * - All the parameters (slots) of tasks are JavaScript objects
 * 
 * - How to fill in my hypothesis:
 *   - function Book(t)   // The JSObject part of it
 *     {
 *        theBook = t;   // WRAPPER for the java object
 *     }
 *     
 *   - within javascript: import package.mypackage.App
 *               App.Gui.m()   <-- give it the class, the class has a static instance. Call that.
 *               
 *   - TaskEngine.eval("....") for updating stuff (GREAT FOR UPDATING DATABASE)
 *   
 * *****************************************************
 * - (1) leave the data in javascript.
 * - (2) when push a swing button, do TaskEngine.eval() or processCommand("done X / Y / A")
 *       - Oh wow! I don't need to do any setting of params for tasks! I just need to do
 *         done / X / Y / A !!!!
 *       - Or you can do execute to ensure that a script is run, therefore I can do my automatic
 *         filling in of a hypothesis!!!
 * - (3) I need to give javascript a copy of my GUI to give it the appropriate stuff.
 * 
 *          
 *          
 * *****************************************************
 * - getFocus() -> return task. You can get all the task parameters
 *    - Have this bound to the current focus for convenience so we can figure out values of parameters
 *    - $this 
 * 
 * - Decomposition and doing a task. It'll fill in the right place as long as the task is live.
 *   If only one decomp available, the tree will be actively decomposed.
 *   
 * - Does not expand a recipe until it is live. Given applicability conditions, ask user which one to choose.
 *   You need to tell it which decomps to do unless you advance focus. Given applicability conditions,
 *   it will attempt to decompose if one is available.
 *   
 *   Me: Do a search thru task tree to see if the person did a step out of order and report as such.
 *       I know what they will do b/c my stuff maps to actions; is the action live?
 * 
 * - Preconds & Postconds are not preactions and postactions. It's a TRUE/FALSE action.  
 * 
 * - Two ways to have things skipped: A task is optional. If a postcondition is fortuitously true don't do it. 
 * sufficient="true" it's the ONLY reason why I'm doing it. A *VERY* strong statement. Do your side effect things
 * in the GROUNDING script. ONLY PRIMITIVES HAVE GROUNDING SCRIPTS.
 * If the user can do it sometimes and the system can do it sometimes, just have a grounding script.
 * 
 * - All sequencing will be done by the task model.
 * 
 * - ADD TASK FOR INCREASING OCCURRENCE OF SOMETHING DONE BY GROUNDING SCRIPT!!!!
 * 
 * - Shoot Guide.println to go the GUI!!!!! 
 * 
 * - I can do a cheezy boolean asserted in a thingie 
 *   to prevent doing the optional CVSCheck 'optional' again.
 * 
 * - ALL KINDS of param string stuff!
 * 
 * 
 * @author Mike Sao Pedro (mikesp@wpi.edu)
 */
public class PedagogicalAgent implements ActionListener, ChangeListener
{
	// --- Made static to simplify enabling access of this object by the
	//     javascript.
	public static InclinedPlaneExperimentComboBoxJPanel mainGui;
	
	// --- The task model file (Assuming you're running this in Eclipse,
	//     the environment is assuming it's run in the "bin" directory, so we need
	//     to pop up a level).
	public static String SCRIPT_URI = "edu/assistment/experiments/pedagogy/Inquiry.xml";
	
	protected TextFieldGuide taskGuide;
	protected boolean begin = false;
		
	// --- Cache of GUI widgets to determine who the ActionListener pertained to
	public static JComboBox leftRampSurfaceComboBox, leftRampAngleComboBox, ballTypeComboBox, runLengthComboBox;
	public static JButton runButton, resetButton;
	public static JButton submitButton;
	public static JButton conclusionsButton;
	public static JTextArea historyText;
	public static JTextField commandText;
	public static JTable trialTable;
	public static JComboBox hypothesisTarget, hypothesisComparator, hypothesisOutcome;
	public static JComboBox conclusionCorrect, conclusionOutcome, conclusionComparator;
	
	public static SimulationController simController;
	public static InclinedPlaneModel simModel;
	
	// --- Map from name of controls to actual controls
	private static HashMap<String, JComponent> controlMap;
	
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

	/**
	 * @param gui Contains some way to get all the widgets displayed to the user
	 */
	public PedagogicalAgent(InclinedPlaneExperimentComboBoxJPanel gui, SimulationController controller, InclinedPlaneModel model)
	{
		mainGui = gui;
		simController = controller;
		simModel = model;
		
		// --- Widgets for controlling experiment properties
		leftRampSurfaceComboBox = mainGui.getInclinedPlaneExperimentJPanel().getLeftRampTypeComboBox();
		leftRampAngleComboBox = mainGui.getInclinedPlaneExperimentJPanel().getLeftRampSteepnessComboBox();
		ballTypeComboBox = mainGui.getInclinedPlaneExperimentJPanel().getBallTypeComboBox();
		runLengthComboBox = mainGui.getInclinedPlaneExperimentJPanel().getRunLengthComboBox();
		
		// --- Widgets for running the experiment
		runButton = mainGui.getSimulationControlJPanel1().getRunButton();
		resetButton = mainGui.getSimulationControlJPanel1().getResetButton();
		
		// --- Pedagogical agent stuff
		historyText = mainGui.getPedagogyPanel().getHistory();
		commandText = mainGui.getPedagogyPanel().getCommands();
		submitButton = mainGui.getPedagogyPanel().getSubmit();
		trialTable = mainGui.getPedagogyPanel().getTrials();
		hypothesisTarget = mainGui.getPedagogyPanel().getTargetVariable();
		hypothesisComparator = mainGui.getPedagogyPanel().getComparator();
		hypothesisOutcome = mainGui.getPedagogyPanel().getOutcomeVariable();
		conclusionComparator = mainGui.getPedagogyPanel().getConclusionComparator();
		conclusionOutcome = mainGui.getPedagogyPanel().getConclusionOutcomeVariable();
		conclusionCorrect = mainGui.getPedagogyPanel().getCorrectness();
		conclusionsButton = mainGui.getPedagogyPanel().getConclusionSubmit();
		
		// --- Initialize the Pedagogical Agent
		registerListening();
		taskGuide = new TextFieldGuide(null, mainGui.getPedagogyPanel());
		
		// --- Get the right URI
		java.net.URL url = Guide.class.getResource("Inquiry.xml");
		PedagogicalAgent.SCRIPT_URI = url.toString();
		System.err.println("Filename = " + PedagogicalAgent.SCRIPT_URI);
		taskGuide.load(PedagogicalAgent.SCRIPT_URI);
		
		// --- Set up control mapping.
		controlMap = new HashMap<String, JComponent>();
		controlMap.put("surface type", leftRampSurfaceComboBox);
		controlMap.put("steepness", leftRampAngleComboBox);
		controlMap.put("ball type", ballTypeComboBox);
		controlMap.put("run length", runLengthComboBox);
		
	}
	 
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

	/**
	 * Listen to all important GUI events so they can be translated to the task model.
	 */
	public void registerListening()
	{
		leftRampAngleComboBox.addActionListener(this);
		leftRampSurfaceComboBox.addActionListener(this);
		ballTypeComboBox.addActionListener(this);
		runLengthComboBox.addActionListener(this);
		runButton.addActionListener(this);
		resetButton.addActionListener(this);
		submitButton.addActionListener(this);
		hypothesisTarget.addActionListener(this);
		hypothesisComparator.addActionListener(this);
		hypothesisOutcome.addActionListener(this);
		conclusionComparator.addActionListener(this);
		conclusionOutcome.addActionListener(this);
		conclusionCorrect.addActionListener(this);
		conclusionsButton.addActionListener(this);
	}
	
	 
	/**
	 * Process when the user hits the 'run' button.
	 * Here is where we let the task engine determine if they
	 * setup the experiment correctly.
	 */
	public void processRunExperiment()
	{
		// --- "Setup" the trial
		//     Order: angle, start, friction, ball
		String angle = leftRampAngleComboBox.getSelectedItem().toString();
		String start = runLengthComboBox.getSelectedItem().toString();
		String friction = leftRampSurfaceComboBox.getSelectedItem().toString();
		String ball = ballTypeComboBox.getSelectedItem().toString();
		
		String query = "SetupTrial / '" + angle + "' / '" + start + "' / '" + friction + "' / '" + ball + "'";
		System.err.println("Sent query to done: " + query);
		taskGuide.done(query);
	}
	
	/**
	 * Advance conversation by automatically building javascript objects, checking
	 * results, speaking to the user, etc. 
	 * 
	 * Note that it uses all 'live' tasks from the toplevel just in case the focus is at a higher-level
	 * task instead of a primitive one. This is sort of a hack for handling cases where the focus
	 * is not on a primitive task in which case we can still advance the conversation if appropriate
	 * given the task model definition. 
	 */
	public void advanceConversation()
	{
		boolean match;
		do
		{
		  match = false;
		  
		  // --- If there is no plan in focus, we are done.
		  Plan focus = taskGuide.engine.getFocus();
		  if(focus == null)
		  {
			  return;
		  }
		  		  
		  List<Plan> list = focus.getLiveDescendants();
		  if(list.isEmpty() && focus.isLive())
		  {
			  list.add(focus);
		  }
		  System.err.println("At: " + focus.getGoal().toString() + "  Current live tasks at focus: " + list.toString());
		  
		  for(Plan p : list)
		  {
			  // --- Automatically advance all those actions which the system can perform
		      if(Guide.isSystem(p))
			  {
				  String id = p.getGoal().getType().getId();
				  //System.err.println("Current goal type match: " + id);
				  
				  match = true;
				  taskGuide.propose(p);
				  taskGuide.execute(p);
				  taskGuide.updateFocus();
				  taskGuide.next("");
				  
				  String isPromptTask = p.getGoal().getProperty("@prompt");
				  if(isPromptTask != null)
				  {
					  //System.err.println("Prompting task: " + top_id);
					  this.respondToUser(p.getGoal().format());
				  }
			  }
		  }
		} while(match == true);		
	}
	
	/**
	 * Process what happens when the user clicks on 'reset'. Essentially this
	 * keeps the task model progressing forward to let correctness processing,
	 * pedagogical actions, etc inside the task model to occur automatically.
	 * 
	 * This is not done off of the 'run' button, because the animation will
	 * not occur due to threading issues. (My guess is because all of these
	 * calls are happening in the GUI thread thus preventing screen repaints from
	 * occurring). 
	 */
	public void processResetExperiment()
	{
		String query = "ResetRamp";
		System.err.println("Sent query to done (process reset experiment): " + query);
		taskGuide.done(query);
		
		advanceConversation();
	}

	public void processConclusions()
	{
		// --- "Setup" the trial
		//     Order: angle, start, friction, ball
		String outcome = conclusionOutcome.getSelectedItem().toString();
		String comparator = conclusionComparator.getSelectedItem().toString();
		String correct = conclusionCorrect.getSelectedItem().toString();
		
		String query = "SelectConclusions / '" + correct + "' / '" + outcome + "' / '" + comparator + "'";
		System.err.println("Sent query to done (process conclusions): " + query);
		taskGuide.done(query);

		advanceConversation();
	}

	/**
	 * Process any text input from the user. I was hoping to have
	 * the student type in answers to some of the questions and use
	 * the language parser to help in that endeavor. Didn't get around to it.
	 * 
	 * @return response from the system
	 */
	 public String processUserUtterance(String utterance)
	 {
		 String response = "??";
		 
		 // --- Handle user requests like 'help' which will suggest
		 //     a next step to perform
		 if(utterance.equals("begin"))
		 {
			 startExperiment(false);
			 response = "Great! Let's begin.";
		 }
		 
		 // --- Last resort: let the guide handle the request.
		 else
		 {
			 try
			 {
				 taskGuide.processCommand(utterance);
				 response = taskGuide.getLastResponse();
			 }
			 catch(Throwable e)
			 {
				 response = "Help, I'm broken!!";
				 System.err.println(e.getLocalizedMessage());
				 e.printStackTrace();
			 }
		 }
		 
		 
		 return response;
	 }
	 
	 public void respondToUser(String response)
	 {
		 this.mainGui.getPedagogyPanel().addToConversation("Mikey", response);
	 }
	 	 
	 /**
	  * If all the pieces of the hypothesis have been filled in, make sure
	  * to advance the task engine to take care of creating objects
	  */
	 public void checkHypothesisAdvance()
	 {		 
		 Plan focus = taskGuide.engine.getFocus();
		 Task top = focus.getGoal();
		 System.err.println("Current goal: " + top.toString());
		 
		 advanceConversation();
	 }
	 
	 /**
	  * Given the name of a GUI object (ie: steepness), highlight it
	  * in red.
	  * 
	  * @param name
	  */
	 public static void highlightControl(String name, java.awt.Color color)
	 {
		JComponent component = controlMap.get(name);
		System.err.println("Highlight: " + name + "  comp: " + component);
		component.setBorder(new LineBorder(color, 2, false));
	 }
	 
	 /**
	  * Clear the highlights of a GUI object
	  * 
	  * @param name
	  */
	 public static void clearHighlight(String name)
	 {
		JComponent component = controlMap.get(name) ;
		component.setBorder(null);
	 }
	 
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
	 
	private void startExperiment(boolean userMakesHypothesis)
	{
		if(begin)
			return;
		
		String option = userMakesHypothesis ? "yes" : "no";
		taskGuide.task("PerformExperiment / '" + option + "'" + " / ");
		taskGuide.next("");
		begin = true;
	}

	/**
	 * Interpret button / widget actions.
	 */
	public void actionPerformed(ActionEvent e) 
	{
		if(e.getSource() == submitButton)
		{
			String command = commandText.getText();
			String response = processUserUtterance(command);
			mainGui.getPedagogyPanel().addToConversation("User", command);
			respondToUser(response);
			commandText.setText("");
			advanceConversation();
		}
		if(e.getSource() == runButton)
		{
			startExperiment(true);
			processRunExperiment();
		}
		if(e.getSource() == conclusionsButton)
		{
			startExperiment(true);
			processConclusions();
		}
		if(e.getSource() == resetButton)
		{
			processResetExperiment();
		}
		if(e.getSource() == hypothesisComparator)
		{
			startExperiment(true);

			// --- Make sure system hasn't done this for us
			Plan p = taskGuide.engine.getTop(taskGuide.engine.getFocus());
			for(Plan d : p.getLiveDescendants())
			{
				if(d.getGoal().getType().getId().equals("SelectComparator"))
				{
					processUserUtterance("done SelectComparator / '" + hypothesisComparator.getSelectedItem().toString() + "'");
				}
			}
			checkHypothesisAdvance();
		}
		if(e.getSource() == hypothesisOutcome)
		{
			startExperiment(true);

			// --- Make sure system hasn't done this for us
			Plan p = taskGuide.engine.getTop(taskGuide.engine.getFocus());
			for(Plan d : p.getLiveDescendants())
			{
				if(d.getGoal().getType().getId().equals("SelectOutcomeVar"))
				{
					processUserUtterance("done SelectOutcomeVar / '" + hypothesisOutcome.getSelectedItem().toString() + "'");
				}
			}
			
			checkHypothesisAdvance();
		}
		if(e.getSource() == hypothesisTarget)
		{
			startExperiment(true);
			
			// --- Make sure system hasn't done this for us
			Plan p = taskGuide.engine.getTop(taskGuide.engine.getFocus());
			for(Plan d : p.getLiveDescendants())
			{
				if(d.getGoal().getType().getId().equals("SelectIndependentVar"))
				{
					processUserUtterance("done SelectIndependentVar / '" + hypothesisTarget.getSelectedItem().toString() + "'");		
				}
			}
			
			checkHypothesisAdvance();
		}
	}

	public void stateChanged(ChangeEvent e) {
		// TODO Auto-generated method stub
		
	}
	 
}
